package login

import (
	"encoding/gob"
	"log"
	"net/http"
	"time"

	"github.com/gorilla/securecookie"
)

func init() {
	gob.Register(&loginSession{})
}

const sessionCookieName = "_sso_auth"

type sessionManager struct {
	sc   *securecookie.SecureCookie
	path string
}

func newSessionManager(path string, authKey, encKey []byte, ttl time.Duration) *sessionManager {
	sc := securecookie.New(authKey, encKey)
	sc.MaxAge(int(ttl.Seconds()))
	// The JSON encoder generates smaller data than gob in our case.
	sc.SetSerializer(&securecookie.JSONEncoder{})
	return &sessionManager{
		sc:   sc,
		path: path,
	}
}

func (m *sessionManager) getSession(req *http.Request) (*loginSession, error) {
	cookie, err := req.Cookie(sessionCookieName)
	if err != nil {
		return nil, err
	}
	var s loginSession
	err = m.sc.Decode(sessionCookieName, cookie.Value, &s)
	if err != nil {
		return nil, err
	}
	return &s, nil
}

func (m *sessionManager) setSession(w http.ResponseWriter, session *loginSession) (err error) {
	var encoded string
	if !session.deleted {
		encoded, err = m.sc.Encode(sessionCookieName, session)
		if err != nil {
			return
		}
	}
	cookie := &http.Cookie{
		Name:     sessionCookieName,
		Value:    encoded,
		Path:     m.path,
		Secure:   true,
		HttpOnly: true,
	}
	if session.deleted {
		cookie.MaxAge = -1
	}
	http.SetCookie(w, cookie)
	return
}

// Wrapper for an http.ResponseWriter that ensures all tracked
// sessions are saved before the request body is sent.
//
// We have to duplicate the logic to call WriteHeader on the first
// Write, otherwise the underlying ResponseWriter won't call our
// WriteHeader function but its own instead.
type sessionResponseWriter struct {
	http.ResponseWriter
	headerWritten bool
	req           *http.Request
	mgr           *sessionManager
	session       *loginSession
}

func (w *sessionResponseWriter) WriteHeader(statusCode int) {
	if statusCode >= 200 && statusCode < 400 {
		if err := w.mgr.setSession(w, w.session); err != nil {
			log.Printf("error saving sessions: %v", err)
		}
	}
	w.ResponseWriter.WriteHeader(statusCode)
	w.headerWritten = true
}

func (w *sessionResponseWriter) Write(b []byte) (int, error) {
	if !w.headerWritten {
		w.WriteHeader(http.StatusOK)
	}
	return w.ResponseWriter.Write(b)
}

// newSessionResponseWriter returns a wrapped http.ResponseWriter that
// will always save the session before writing the response body.
func (m *sessionManager) newSessionResponseWriter(w http.ResponseWriter, req *http.Request, session *loginSession) http.ResponseWriter {
	return &sessionResponseWriter{
		ResponseWriter: w,
		req:            req,
		mgr:            m,
		session:        session,
	}
}
